import hashlib
import json
import sys
from datetime import UTC, datetime, timedelta
from queue import Queue
from typing import Generator, Optional
from uuid import UUID, uuid4

import click

from highlighter.agent.agent import HLAgent, set_mock_aiko_messager
from highlighter.agent.capabilities.base_capability import DataSourceType
from highlighter.client.base_models.data_file import DataFile
from highlighter.client.gql_client import HLClient
from highlighter.client.tasks import (
    Task,
    TaskStatus,
    lease_task,
    lease_tasks_from_steps,
    update_task,
    update_task_status,
)
from highlighter.core.logging import ColourStr


def _is_uuid(s):
    try:
        _ = UUID(s)
        return True
    except Exception as _:
        return False


def iter_buffer(
    buffer,
    sep: bytes,
) -> Generator[bytes, None, None]:
    """
    Streams data from stdin until the Nth occurrence of a separator byte sequence.

    sep: The separator byte sequence (e.g., b'\n' or b'\x00\x01')
    """
    if sep is None or len(sep) == 0:
        raise ValueError("Separator must be a non-empty bytes object")

    byte_result = bytearray()
    buffer_remainder = bytearray()

    while True:
        chunk = buffer.read(4096)  # Adjusted for testing, increase for performance
        if not chunk:
            if byte_result:
                byte_result = bytes(byte_result)
                yield byte_result
            break

        buffer_remainder.extend(chunk)

        while True:
            sep_index = buffer_remainder.find(sep)
            if sep_index == -1:
                break

            byte_result.extend(buffer_remainder[:sep_index])
            yield bytes(byte_result)

            byte_result.clear()
            buffer_remainder = buffer_remainder[sep_index + len(sep) :]

        byte_result.extend(buffer_remainder)
        buffer_remainder.clear()

    if byte_result:
        yield bytes(byte_result)


@click.group("agent")
@click.pass_context
def agent_group(ctx):
    pass


# ToDo: Now that I am not accepting cli passed stream params. Is this needed,
# or is it just handled by the aiko pipeline
def parse_stream_params(agent, agent_definition) -> dict:

    stream_params = {}
    for node in agent.pipeline_graph.nodes():
        node_name = node.name

        # Start with in code params
        default_stream_parameters = node.element.default_stream_parameters()

        # Overwite with global pipeline definition params
        global_pipeline_definition_params = {
            k: v
            for k, v in agent_definition.parameters
            if k in node.element.DefaultStreamParameters.model_fields
        }
        default_stream_parameters.update(global_pipeline_definition_params)

        # Overwite with per element pipeline definition paras
        element_definition = [e for e in agent_definition.elements if e.name == node_name][0]
        pipeline_element_definition_params = {
            k.replace(f"{node_name}.", ""): v
            for k, v in element_definition.parameters.items()
            if k.replace(f"{node_name}.", "") in node.element.DefaultStreamParameters.model_fields
        }
        node.element.parameters = pipeline_element_definition_params
        default_stream_parameters.update(pipeline_element_definition_params)

        ele_stream_params = node.element.DefaultStreamParameters(**default_stream_parameters).model_dump()

        stream_params.update({f"{element_definition.name}.{k}": v for k, v in ele_stream_params.items()})

    return stream_params


def iter_queue(q):
    while not q.empty():
        return q.get()


class OnAgentError:
    def __init__(self, task_id=None):
        self.task_id = task_id

    def __call__(self, agent):
        print(ColourStr.red(f"Task {self.task_id} FAILED: {agent.hl_task_status_info}"))
        if self.task_id is not None:
            client = HLClient.get_client()
            update_task_status(
                client,
                str(self.task_id),
                TaskStatus.FAILED,
                message=json.dumps(agent.hl_task_status_info["message"]),
            )


class OnAgentStop:
    def __init__(self, task_id=None):
        self.task_id = task_id

    def __call__(self, agent):
        print(ColourStr.green(f"Task {self.task_id} SUCCESS: {agent.hl_task_status_info}"))
        if self.task_id is not None:
            client = HLClient.get_client()
            update_task_status(
                client,
                str(self.task_id),
                TaskStatus.SUCCESS,
                message=json.dumps(agent.hl_task_status_info["message"]),
            )


INITIAL_LEASE_TIME = 30 * 60  # 30min
LEASE_TIME_UPDATE_DELTA = 30 * 60  # 30min
TIME_LEFT_BEFORE_UPDATING = 5 * 60  # 5min


class OnBeforeProcessFrameUpdateTaskLease:
    def __init__(self, task_id, update_delta: int, time_left_before_updating: int):
        self.task_id = task_id
        self.update_delta = update_delta  # seconds
        self.time_left_before_updating = time_left_before_updating  # seconds
        self._client = HLClient.get_client()

    def __call__(self, agent):

        task_leased_until = self._client.task(return_type=Task, id=self.task_id).leased_until
        if task_leased_until is None:
            task_leased_until = datetime.now(UTC)
        else:
            task_leased_until = task_leased_until

        sec_remaining = (task_leased_until - datetime.now(UTC)).total_seconds()
        agent.logger.info(f"task {self.task_id} has {sec_remaining} seconds remaining on lease")

        if sec_remaining < self.time_left_before_updating:
            new_leased_until = task_leased_until + timedelta(seconds=self.update_delta)
            agent.logger.info(f"update {self.task_id} from {task_leased_until} to {new_leased_until}")
            update_task(self._client, self.task_id, status=TaskStatus.RUNNING, leased_until=new_leased_until)


def _get_next_task_data_sources_from_step_id(step_id, client):
    step_id = UUID(step_id)

    def get_next():
        while True:
            tasks = lease_tasks_from_steps(
                client,
                [step_id],
                lease_sec=3600,
                filter_by_status="PENDING",
                set_status_to="RUNNING",
                count=1,
            )
            if not tasks:
                return None

            task = tasks[0]
            file_urls = [d.file_url_original for d in task.case.latest_submission.data_files]
            file_ids = [d.uuid for d in task.case.latest_submission.data_files]
            media_types = [d.content_type for d in task.case.latest_submission.data_files]

            data_sources = [
                DataSourceType(media_type=m, url=u, id=i) for m, u, i in zip(media_types, file_urls, file_ids)
            ]
            yield task.id, data_sources

    return get_next()


def _get_next_task_data_sources_from_step_task_ids(step_task_ids, client):
    step_task_ids = step_task_ids.split(",")
    assert all([_is_uuid(i) for i in step_task_ids])

    def get_next():
        for task_id in step_task_ids:
            task = lease_task(
                client,
                task_id=task_id,
                lease_sec=3600,
                set_status_to="RUNNING",
            )
            file_urls = [d.file_url_original for d in task.case.latest_submission.data_files]
            file_ids = [d.uuid for d in task.case.latest_submission.data_files]
            media_types = [d.content_type for d in task.case.latest_submission.data_files]

            data_sources = [
                DataSourceType(media_type=m, url=u, id=i) for m, u, i in zip(media_types, file_urls, file_ids)
            ]
            yield task.id, data_sources
        return None

    return get_next()


def _get_next_task_data_sources_from_filepaths(input_data, seperator):
    if input_data == "--":
        file_paths = iter([f.decode("utf-8") for f in iter_buffer(sys.stdin.buffer, sep=seperator)])
    else:
        file_paths = input_data.split(seperator)

    def get_next():
        for i, filepath in enumerate(file_paths):
            data_source = DataSourceType(media_type="", url=filepath, id=UUID(int=i))
            yield None, [data_source]

    return get_next()


def _get_next_task_data_sources_from_stdin_raw_data(input_data, seperator):
    def get_next():
        for i, content in enumerate(iter_buffer(sys.stdin.buffer, sep=seperator)):
            data_source = DataSourceType(media_type="", url="bytes://", id=UUID(int=i), content=content)
            yield None, [data_source]

    return get_next()


def _reading_filepaths_from_stdin(input_data, expect_filepaths):
    return (input_data == "--") and (not sys.stdin.isatty() and (expect_filepaths))


def _reading_raw_data_from_stdin(input_data, expect_filepaths):
    return (input_data == "--") and (not sys.stdin.isatty() and (not expect_filepaths))


def _reading_filepaths_from_input_data(input_data, expect_filepaths):
    return isinstance(input_data, str) and (input_data != "--") and (expect_filepaths)


def loop_over_task_data_sources(agent, stream_id, input_name, get_next_task_iter, queue_response):

    for i, (task_id, data_sources) in enumerate(get_next_task_iter):

        stream_parameters: dict = parse_stream_params(
            agent.pipeline,
            agent.pipeline_definition,
        )
        stream_parameters[input_name] = data_sources

        if task_id is not None:
            agent.set_callback("on_agent_error", OnAgentError(task_id=task_id))
            agent.set_callback("on_agent_stop", OnAgentStop(task_id=task_id))
            agent.set_callback(
                "on_before_process_frame",
                OnBeforeProcessFrameUpdateTaskLease(
                    task_id,
                    LEASE_TIME_UPDATE_DELTA,
                    TIME_LEFT_BEFORE_UPDATING,
                ),
            )
            _stream_id = task_id
        else:
            _stream_id = UUID(fields=(stream_id, 0, 0, 0, 0, i))

        agent.run(
            _stream_id,
            stream_parameters,
            mqtt_connection_required=False,
            queue_response=queue_response,
        )
        agent.pipeline.destroy_stream(_stream_id)


def loop_over_process_frame(agent, stream_id, frame_datas, queue_response):
    # This function can be removed once
    # the issue it's solving is resolved.
    # See to function's doc string for more info
    set_mock_aiko_messager()

    if isinstance(frame_datas, dict):
        frame_datas = [frame_datas]

    stream_parameters: dict = parse_stream_params(
        agent.pipeline,
        agent.pipeline_definition,
    )

    agent.pipeline.create_stream(stream_id, parameters=stream_parameters, queue_response=queue_response)
    for frame_id, frame in enumerate(frame_datas):
        stream = {
            "stream_id": stream_id,
            "frame_id": frame_id,
        }

        data_files = [
            DataFile(
                file_id=uuid4(),
                content=frame["content"],
                media_frame_index=0,
                content_type="text",
            )
        ]
        agent.pipeline.process_frame(stream, {"data_files": data_files})
    agent.pipeline.destroy_stream(stream_id)


DEFAULT_FILEPATH_SEPERATOR = "\n".encode("utf-8")
DEFAULT_CONTENT_SEPERATOR = "===END==".encode("utf-8")


@agent_group.command("run")
@click.option(
    "--buffer-seperator",
    "-p",
    type=str,
    default=None,
    help="If --expect-filepaths is set the default is '\\n'. Else the the unix file seperator '\\x1c'. This parameter is only used for piped inputs, if passing paths directly use spaces to separate paths",
)
@click.option("--expect-filepaths", "-f", is_flag=True, default=False)
@click.option("--step-task-ids", "-t", type=str, default=None)
@click.option("--step-id", "-i", type=str, default=None)
@click.option("--stream-id", "-s", type=int, default=1)
@click.option("--dump-definition", type=str, default=None)
@click.option("--input-name", type=str, default="Source.data_sources")
@click.argument("agent_definition", type=click.Path(dir_okay=False, exists=False))
@click.argument("input_data", type=click.Path(exists=False), default="--", required=False)
@click.pass_context
def _run(
    ctx,
    buffer_seperator,
    expect_filepaths,
    step_task_ids,
    step_id,
    stream_id,
    dump_definition,
    input_name,
    agent_definition,
    input_data,
):
    """Run a local Highlighter Agent to process data either on your local machine or as Highlighter Tasks.

    When processing local files, a mock task is constructed containing a single
    data file (e.g., image, text, video). Each file results in one stream. For
    example, if you're processing a set of images, each image will create a new
    stream, with each stream processing a single frame. In the case of a video,
    a stream will be created for each video, and each stream will process every
    frame of its respective video.

    Similarly, when processing Highlighter tasks, a stream is created for each
    task, and each stream processes all the data within that task. For instance,
    if a task contains several images, all the images in that task will be
    processed within the corresponding stream.

    The Agent definition must have its first element as a
    `DataSourceCapability`, such as `ImageDataSource`, `VideoDataSource`,
    `TextDataSource`, `JsonArrayDataSource`, etc. The following examples assume
    the use of `ImageDataSource`.

    Examples:

      \b
      1. Run an agent against a single image path
      \b
        > hl agent run -f agent-def.json images/123.jpg

      \b
      2. Run an agent against a multiple image paths
      \b
        > find -name *.jpg images/ | hl agent run -f agent-def.json

      \b
      3. Cat the contents of an image to an agent
      \b
        > images/123.jpg | hl agent run -f agent-def.json

      \b
      4. Pass data directly to process_frame
      \b
        > hl agent run -f agent-def.json '[{"foo": "bar"},{"foo": "baz"}]'

    """
    if step_id and step_task_ids:
        raise ValueError()

    client = ctx.obj["client"]
    queue_response = ctx.obj.get("queue_response", None)
    if step_id:

        get_next_task_iter = _get_next_task_data_sources_from_step_id(step_id, client)

    elif step_task_ids:

        get_next_task_iter = _get_next_task_data_sources_from_step_task_ids(step_task_ids, client)

    elif _reading_raw_data_from_stdin(input_data, expect_filepaths):

        if buffer_seperator is None:
            seperator = DEFAULT_CONTENT_SEPERATOR
        else:
            seperator = buffer_seperator.encode("utf-8")

        get_next_task_iter = _get_next_task_data_sources_from_stdin_raw_data(input_data, seperator)

    elif _reading_filepaths_from_stdin(input_data, expect_filepaths):

        if buffer_seperator is None:
            seperator = DEFAULT_FILEPATH_SEPERATOR
        else:
            seperator = buffer_seperator.encode("utf-8")

        get_next_task_iter = _get_next_task_data_sources_from_filepaths(input_data, seperator)

    elif _reading_filepaths_from_input_data(input_data, expect_filepaths):

        get_next_task_iter = _get_next_task_data_sources_from_filepaths(input_data, ",")

    else:
        # assume process_frame_data is passed in directly either as a json
        # str in input_data or via sdtin buffer
        # assert False, f"-------------: {input_data}"
        try:
            if input_data == "--":
                frame_datas = json.load(sys.stdin.buffer)
            else:
                frame_datas = json.loads(input_data)
        except Exception as e:
            raise ValueError(f"{e} -- {input_data}")

        get_next_task_iter = None

    agent = HLAgent(agent_definition, dump_definition=dump_definition)

    if get_next_task_iter is None:
        loop_over_process_frame(agent, stream_id, frame_datas, queue_response)
    else:
        loop_over_task_data_sources(agent, stream_id, input_name, get_next_task_iter, queue_response)

    # for task in tasks:
    #    assessment_uuid = task.case.latest_submission.uuid
    #    endpoint_url = HLClient.get_client().endpoint_url
    #    endpoint_url = endpoint_url.replace("graphql", f"oid/assessment/{assessment_uuid}")
    #    data_sources = {
    #        "HlAssessmentRead": {
    #            "media_type": "highlighter-assessment",
    #            "url": endpoint_url,
    #        }
    #    }

    #    _stream_parameters: dict = parse_stream_params(
    #        agent.pipeline,
    #        agent.pipeline_definition,
    #        stream_parameters,
    #        params={"data_sources": data_sources},
    #    )

    #    stream_id = task.id
    #    agent.run(
    #        stream_id,
    #        _stream_parameters,
    #        mqtt_connection_required=False,
    #    )



# Notes
# ~~~~~
# This is a low-level mechanism, used by the AikoServices framework
# and should not have any static dependencies on that framework
#
# To Do
# ~~~~~
# - Add support for ConnectionState.TRANSPORT (MQTT) connection state updates

__all__ = ["ConnectionState", "Connection"]

class ConnectionState:
    NONE = "NONE"
    NETWORK = "NETWORK"      # Wi-Fi or Ethernet available
    BOOTSTRAP = "BOOTSTRAP"  # MQTT configuration found
    TRANSPORT = "TRANSPORT"  # MQTT, Ray, ROS2, etc
    REGISTRAR = "REGISTRAR"  # Registrar available for use

    states = [NONE, NETWORK, TRANSPORT, REGISTRAR]  # order matters

    @classmethod
    def index(cls, connection_state):  # throws ValueError
        return cls.states.index(connection_state)

class Connection:
    def __init__(self):
        self.connection_state = ConnectionState.NONE
        self.connection_state_handlers = []

    def add_handler(self, connection_state_handler):
        connection_state_handler(self, self.connection_state)
        if not connection_state_handler in self.connection_state_handlers:
            self.connection_state_handlers.append(connection_state_handler)

    def is_connected(self, connection_state):  # throws ValueError
        current_state_index = ConnectionState.index(self.connection_state)
        return current_state_index >= ConnectionState.index(connection_state)

    def remove_handler(self, connection_state_handler):
        if connection_state_handler in self.connection_state_handlers:
           self.connection_state_handlers.remove(connection_state_handler)

    def update_state(self, connection_state):
        self.connection_state = connection_state
        for connection_state_handler in self.connection_state_handlers:
            connection_state_handler(self, connection_state)



import shutil
from copy import deepcopy
from pathlib import Path

import ultralytics
import yaml

__all__ = ["generate", "prepare_datasets", "train"]


def generate(training_run_dir: Path):
    shutil.copy(Path(__file__).parent / "template.py", training_run_dir / "trainer.py")

    default_cfg = dict(ultralytics.cfg.get_cfg())
    default_cfg["model"] = "yolov8m.pt"
    default_cfg["project"] = "runs"
    with (training_run_dir / "cfg.yaml").open("w") as f:
        yaml.dump(default_cfg, f)


def prepare_datasets(datasets):
    # When creating a training run in Highlighter the train split is required
    # but a user can supply either a test or dev set, or both. If not both we
    # duplicate the one that exists here
    if "test" not in datasets:
        datasets["test"] = deepcopy(datasets["dev"])
        datasets["test"].data_files_df.split = "test"
    if "dev" not in datasets:
        datasets["dev"] = deepcopy(datasets["test"])
        datasets["dev"].data_files_df.split = "dev"

    # Combine the Highlighter Datasets together because this is what the YoloWriter
    # expects
    combined_ds = datasets["train"]
    combined_ds.append([datasets["dev"], datasets["test"]])

    # Ultralytics name their dataset splits differently so
    # we need to map them
    #   their "val" is our "test"
    #   their "test" is our "dev"
    combined_ds.data_files_df.loc[combined_ds.data_files_df.split == "test", "split"] = "val"
    combined_ds.data_files_df.loc[combined_ds.data_files_df.split == "dev", "split"] = "test"
    return combined_ds


def train(cfg, training_run_dir: Path):
    model = ultralytics.YOLO(cfg["model"])

    ultralytics.settings.update({"datasets_dir": str(training_run_dir.absolute())})

    data_cfg_path = (training_run_dir / "datasets" / "data.yaml").absolute()
    with data_cfg_path.open("r") as f:
        data_cfg = yaml.safe_load(f)

    with data_cfg_path.open("w") as f:
        yaml.dump(data_cfg, f)

    cfg["data"] = str(data_cfg_path)

    model.train(**cfg)
    return model



import json
import os
from pathlib import Path
from tempfile import mkdtemp
from typing import Callable, Dict, Optional, Tuple, Union
from uuid import UUID

import aiko_services as aiko
import yaml
from aiko_services.main import aiko as aiko_main

__all__ = [
    "HLAgent",
    "set_mock_aiko_messager",
]


def load_pipeline_definition(path) -> Dict:
    path = Path(path)
    suffix = path.suffix

    if suffix in (".json",):
        with path.open("r") as f:
            pipeline_def = json.load(f)
    elif suffix in (".yml", ".yaml"):
        with path.open("r") as f:
            pipeline_def = yaml.safe_load(f)
    else:
        raise NotImplementedError(
            f"Unsupported pipeline_definition file, '{path}'." " Expected .json|.yml|.yaml"
        )

    def remove_dict_keys_starting_with_a_hash(data):
        if isinstance(data, dict):
            # Create a new dictionary excluding keys starting with "#"
            return {
                key: remove_dict_keys_starting_with_a_hash(value)
                for key, value in data.items()
                if not key.startswith("#")
            }
        elif isinstance(data, list):
            # If the item is a list, recursively apply the function to each element
            return [remove_dict_keys_starting_with_a_hash(item) for item in data]
        else:
            # If the item is neither a dict nor a list, return it as-is
            return data

    pipeline_def = remove_dict_keys_starting_with_a_hash(pipeline_def)
    return pipeline_def


class HLPipelineImpl(aiko.PipelineImpl):

    def __init__(self, context):
        self.hl_task_status_info = {
            "state": aiko.StreamState.RUN,
            "message": "",
        }
        self.on_agent_error: Optional[Callable] = None
        self.on_agent_stop: Optional[Callable] = None
        self.on_before_process_frame: Optional[Callable] = None
        self.on_after_process_frame: Optional[Callable] = None
        super().__init__(context)

    def _process_stream_event(self, element_name, stream, stream_event, diagnostic, in_destroy_stream=False):
        # class StreamState:
        #    ERROR =   -2  # Don't generate new frames and ignore queued frames
        #    STOP  =   -1  # Don't generate new frames and process queued frames
        #    RUN   =    0  # Generate new frames and process queued frames

        most_recent_state = self.hl_task_status_info["state"]

        if stream_event == aiko.StreamEvent.ERROR:
            if aiko.StreamState.ERROR < most_recent_state:
                self.hl_task_status_info["state"] = aiko.StreamState.ERROR
                self.hl_task_status_info["message"] = diagnostic

        elif stream_event == aiko.StreamEvent.STOP:
            if aiko.StreamState.STOP < most_recent_state:
                self.hl_task_status_info["state"] = aiko.StreamState.STOP
                self.hl_task_status_info["message"] = diagnostic

        return super()._process_stream_event(
            element_name, stream, stream_event, diagnostic, in_destroy_stream=in_destroy_stream
        )

    def run_pipeline_with_callabcks(self, *args, **kwargs):
        self.run(*args, **kwargs)

        hl_task_status = self.hl_task_status_info

        if (hl_task_status["state"] == aiko.StreamState.ERROR) and self.on_agent_error is not None:
            self.on_agent_error(self)
        elif (hl_task_status["state"] == aiko.StreamState.STOP) and self.on_agent_stop is not None:
            self.on_agent_stop(self)

    def process_frame(self, stream_dict, frame_data) -> Tuple[aiko.StreamEvent, dict]:
        try:
            if self.on_before_process_frame is not None:
                self.on_before_process_frame(self)
        except Exception as e:
            return aiko.StreamEvent.ERROR, {"diagnostic": e}

        result = self._process_frame_common(stream_dict, frame_data, True)

        try:
            if self.on_after_process_frame is not None:
                self.on_after_process_frame(self)
        except Exception as e:
            return aiko.StreamEvent.ERROR, {"diagnostic": e}

        return result


aiko.PipelineImpl = HLPipelineImpl


def _validate_uuid(s):
    try:
        u = UUID(s)
        return u
    except Exception as e:
        return None


def _validate_path(s) -> Optional[Path]:
    try:
        p = Path(s)
        if p.exists():
            return p
    except Exception as e:
        return None


class HLAgent:

    def __init__(
        self,
        pipeline_definition: Union[str, dict, os.PathLike],
        name: Optional[str] = None,
        dump_definition: Optional[os.PathLike] = None,
    ):
        if dump_definition is not None:
            pipeline_path = Path(dump_definition)
        else:
            pipeline_path = Path(mkdtemp()) / "pipeline_def.json"

        if pipeline_definition_path := _validate_path(pipeline_definition):
            definition_dict = load_pipeline_definition(pipeline_definition_path)
            if name is None:
                name = pipeline_definition_path.name

        elif def_uuid := _validate_uuid(pipeline_definition):
            raise NotImplementedError("Need to implement pulling the pipeline_definition from Highlighter")
        elif isinstance(pipeline_definition, dict):
            if name is None:
                raise ValueError(
                    "If pipeline_definition is a dict you must provide the 'name' arg to HLAgent.__init__"
                )
            definition_dict = pipeline_definition

        else:
            if Path(pipeline_definition).suffix not in (".json", ".yml", ".yaml"):
                raise SystemExit(f"pipeline_definition '{pipeline_definition}' path does not exist")
            else:
                raise SystemExit(f"pipeline_definition '{pipeline_definition}' id does not exist")

        self._dump_definition(definition_dict, pipeline_path)

        parsed_definition = aiko.PipelineImpl.parse_pipeline_definition(pipeline_path)

        init_args = aiko.pipeline_args(
            name,
            protocol=aiko.PROTOCOL_PIPELINE,
            definition=parsed_definition,
            definition_pathname=pipeline_path,
        )
        pipeline = aiko.compose_instance(aiko.PipelineImpl, init_args)

        self.pipeline = pipeline
        self.pipeline_definition = parsed_definition

    def _dump_definition(self, pipeline_def: Dict, path: Path):
        with path.open("w") as f:
            json.dump(pipeline_def, f, sort_keys=True, indent=2)

    def run(self, stream_id, stream_parameters, mqtt_connection_required=False, queue_response=None):
        self.pipeline.create_stream(stream_id, parameters=stream_parameters, queue_response=queue_response)
        self.pipeline.run_pipeline_with_callabcks(mqtt_connection_required=mqtt_connection_required)

    def get_head_data_source_capability(self):
        head_capability = [x for x in self.pipeline.pipeline_graph][0]
        if hasattr(head_capability.element, "_is_data_source"):
            return head_capability
        return None

    def set_callback(self, callback_name: str, callback: Callable):
        setattr(self.pipeline, callback_name, callback)

    def process_frame(self, frame_data, stream_id=0, frame_id=0):
        stream = {
            "stream_id": stream_id,
            "frame_id": frame_id,
        }
        return self.pipeline.process_frame(stream, frame_data)


def set_mock_aiko_messager():
    # ToDo: Chat with Andy about if this is a requirement. The issue is
    # in pipeline.py +999 causes an error because if I use `process_frame`
    # directly, without setting the aiko.message object to something I
    # get an attribute error when .publish is called
    class MockMessage:
        def publish(self, *args, **kwargs):
            pass

        def subscribe(self, *args, **kwargs):
            pass

        def unsubscribe(self, *args, **kwargs):
            pass

    aiko_main.message = MockMessage()



import json
import os
import tempfile
from datetime import datetime, timedelta
from enum import Enum
from typing import Optional
from urllib.parse import urlparse

try:
    import cv2
except ModuleNotFoundError as _:
    cv2 = None

import numpy as np
from PIL import Image

from highlighter.client import download_bytes
from highlighter.client.base_models import DataFile
from highlighter.client.io import _pil_open_image_bytes, _pil_open_image_path
from highlighter.core.exceptions import require_package

from .base_capability import DataSourceCapability, DataSourceType

__all__ = [
    "ImageDataSource",
    "TextDataSource",
    "JsonArrayDataSource",
    "VideoDataSource",
]


class TextFrameIterator:

    def __init__(self, data_source: DataSourceType, byte_encoding="utf-8"):
        self.byte_encoding = byte_encoding
        if data_source.url.startswith("bytes"):
            self._read_text = lambda ds: ds.content.decode(self.byte_encoding)
        elif os.path.isfile(data_source.url):

            def read_text(p):
                with open(p, "r") as f:
                    return f.read()

            self._read_text = lambda ds: read_text(ds.url)
        elif all([urlparse(data_source.url), urlparse(data_source.url).netloc]):
            self._read_text = lambda ds: download_bytes(ds.url).decode(self.byte_encoding)
        else:
            raise ValueError(f"Invalid DataSource.url, expected local_path or url, got: {data_source.url}")

        self.ds = data_source
        self._complete = False

    def __iter__(self):
        return self

    def __next__(self):
        if self._complete:
            raise StopIteration
        self._complete = True
        img = self._read_text(self.ds)
        data_file = DataFile(
            file_id=self.ds.id,
            content=img,
            content_type="text",
            media_frame_index=0,
        )
        return {"data_files": data_file, "entities": {}}


class TextDataSource(DataSourceCapability):
    """

    TODO: Check/update this

    Example:
        # process a single string
        hl agent run --data-source TextDataSource PIPELINE.json "tell me a joke."

        # process many text files
        ToDo

        # Read from stdin
        cat file | hl agent run --data-source TextDataSource -sp read_stdin=true PIPELINE.json
    """

    stream_media_type = "text"

    class DefaultStreamParameters(DataSourceCapability.DefaultStreamParameters):
        byte_encoding: Optional[str] = "utf-8"

    @property
    def byte_encoding(self) -> str:
        value, _ = self._get_parameter("byte_encoding")
        return value

    def frame_data_generator(self, data_sources):
        for ds in data_sources:
            for frame_data in TextFrameIterator(ds, self.byte_encoding):
                yield frame_data


class JsonArrayFrameIterator:
    def __init__(self, data_source: DataSourceType, key: str, byte_encoding="utf-8"):
        self.byte_encoding = byte_encoding

        if data_source.url.startswith("bytes"):
            _json = json.loads(data_source.content.decode(self.byte_encoding))
        elif os.path.isfile(data_source.url):
            with open(data_source.url, "r") as f:
                _json = json.load(f)
        elif all([urlparse(data_source.url), urlparse(data_source.url).netloc]):
            _json = json.loads(download_bytes(data_source.url).decode(self.byte_encoding))
        else:
            raise ValueError(f"Invalid DataSource.url, expected local_path or url, got: {data_source.url}")

        if key:
            for k in key.split("."):
                _json = _json[k]

        self._json_arr = iter([(data, i) for i, data in enumerate(_json)])

        self.ds = data_source
        self._complete = False

    def __iter__(self):
        return self

    def __next__(self):
        try:
            content, media_frame_index = next(self._json_arr)
            data_file = DataFile(
                file_id=self.ds.id,
                content=content,
                content_type="text",
                media_frame_index=media_frame_index,
            )
            return {"data_files": data_file, "entities": {}}
        except StopIteration:
            raise StopIteration


class JsonArrayDataSource(DataSourceCapability):

    stream_media_type = "text"

    class DefaultStreamParameters(DataSourceCapability.DefaultStreamParameters):
        key: str = ""

    @property
    def key(self) -> str:
        value, _ = self._get_parameter("key")
        return value

    def frame_data_generator(self, data_sources):
        for ds in data_sources:
            for frame_data in JsonArrayFrameIterator(ds, self.key):
                yield frame_data


class OutputType(str, Enum):
    numpy = "numpy"
    pillow = "pillow"


class ImageFrameIterator:
    def __init__(self, data_source: DataSourceType, output_type: OutputType):
        self.output_type = output_type
        if data_source.url.startswith("bytes"):
            self._read_image = lambda ds: _pil_open_image_bytes(ds.content)
        elif os.path.isfile(data_source.url):
            self._read_image = lambda ds: _pil_open_image_path(ds.url)
        elif all([urlparse(data_source.url), urlparse(data_source.url).netloc]):

            def _dl_pil_image(ds: DataSourceType):
                image_bytes = download_bytes(ds.url)
                assert image_bytes is not None
                image = _pil_open_image_bytes(image_bytes)
                return image

            self._read_image = _dl_pil_image
        elif data_source.url.startswith("hl-data-file-id"):
            raise NotImplementedError()
        else:
            raise ValueError(f"Invalid DataSource.url, expected local_path or url, got: {data_source.url}")

        self.ds = data_source
        self._complete = False

    def __iter__(self):
        return self

    def __next__(self):
        if self._complete:
            raise StopIteration

        self._complete = True
        img = self._read_image(self.ds)
        if self.output_type == OutputType.numpy:
            img = np.array(img, dtype=np.uint8)

        data_file = DataFile(
            file_id=self.ds.id,
            content=img,
            content_type="image",
            media_frame_index=0,
        )
        return {"data_files": data_file, "entities": {}}


class ImageDataSource(DataSourceCapability):
    """

    Example:
        # process a single image
        hl agent run PIPELINE.json image.jpg

        # process many images
        find image/dir/ -n "*.jpg" | hl agent run PIPELINE.json
    """

    stream_media_type = "image"

    class DefaultStreamParameters(DataSourceCapability.DefaultStreamParameters):
        output_type: OutputType = OutputType.numpy

    @property
    def output_type(self) -> OutputType:
        value, _ = self._get_parameter("output_type")
        return value

    def frame_data_generator(self, data_sources):
        for ds in data_sources:
            for frame_data in ImageFrameIterator(ds, self.output_type):
                yield frame_data


class _AutoDeletingTempFile:
    def __init__(self, content_bytes):
        self.file = tempfile.NamedTemporaryFile(delete=False)
        self.file_path = self.file.name
        self.file.write(content_bytes)
        self.file.flush()
        del content_bytes

    def __del__(self):
        if self.file_path and os.path.exists(self.file_path):
            os.remove(self.file_path)


@require_package(cv2, "cv2", "opencv")
class VideoFrameIterator:
    def __init__(self, data_source: DataSourceType, output_type: OutputType):

        self.output_type = output_type

        if data_source.url.startswith("bytes"):
            self._tmp_file = _AutoDeletingTempFile(data_source.content)
            video_path = self._tmp_file.file_path
        elif os.path.isfile(data_source.url):
            video_path = data_source.url
        elif all([urlparse(data_source.url), urlparse(data_source.url).netloc]):
            self._tmp_file = _AutoDeletingTempFile(download_bytes(data_source.url))
            video_path = self._tmp_file.file_path

        elif data_source.url.startswith("hl-data-file-id"):
            raise NotImplementedError()
        else:
            raise ValueError(f"Invalid DataSource.url, expected local_path or url, got: {data_source.url}")

        self.cap = cv2.VideoCapture(str(video_path))
        if not self.cap.isOpened():
            raise ValueError(f"Cannot open video file: {video_path}")
        self.video_path = video_path
        self.frame_index = 0
        self.ds = data_source
        self.start_time = datetime.now()

    def __iter__(self):
        return self

    def __next__(self):
        ret, frame_img = self.cap.read()
        if not ret:  # No more frames to read
            self.cap.release()
            raise StopIteration

        frame_img = cv2.cvtColor(frame_img, cv2.COLOR_BGR2RGB)

        if self.output_type == OutputType.pillow:
            frame_img = Image.fromarray(frame_img)
        timestamp = self.cap.get(cv2.CAP_PROP_POS_MSEC)
        data_file = DataFile(
            file_id=self.ds.id,
            content=frame_img,
            content_type="image",
            recorded_at=self.start_time + timedelta(milliseconds=timestamp),
            media_frame_index=self.frame_index,
        )
        self.frame_index += 1
        return {"data_files": data_file, "entities": {}}

    def __del__(self):
        if self.cap.isOpened():
            self.cap.release()
        if hasattr(self, "_tmp_file"):
            del self._tmp_file


class VideoDataSource(DataSourceCapability):

    stream_media_type = "video"

    class DefaultStreamParameters(DataSourceCapability.DefaultStreamParameters):
        output_type: OutputType = OutputType.numpy

    @property
    def output_type(self) -> OutputType:
        value, _ = self._get_parameter("output_type")
        return value

    def frame_data_generator(self, data_sources):
        for ds in data_sources:
            for frame_data in VideoFrameIterator(ds, self.output_type):
                yield frame_data



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# export AIKO_LOG_LEVEL=DEBUG
# ./storage.py start
#
# ./storage.py test_command
#   Command: test_command(hello)
# ./storage.py test_request request_0
#   Response: [("request_0", [])]
#
# To Do
# ~~~~~
# - Compare "do_command()" and "do_request()" with HL AM / HL SSM
#   - Refactor into "service.py" ?
#
# - Consider GraphQL over MQTT !

from abc import abstractmethod
import click
import sqlite3

from aiko_services.main import *
from aiko_services.main.transport import *
from aiko_services.main.utilities import *

_VERSION = 0

ACTOR_TYPE = "storage"
PROTOCOL = f"{SERVICE_PROTOCOL_AIKO}/{ACTOR_TYPE}:{_VERSION}"

_LOGGER = aiko.logger(__name__)
_TOPIC_RESPONSE = f"{aiko.topic_out}/storage_response"

# --------------------------------------------------------------------------- #

class Storage(Actor):
    Interface.default("Storage", "aiko_services.main.storage.StorageImpl")

    @abstractmethod
    def test_command(self, parameter):
        pass

    @abstractmethod
    def test_request(self, topic_path_response, request):
        pass

class StorageImpl(Storage):
    def __init__(self, context, database_pathname):
        context.get_implementation("Actor").__init__(self, context)

        self.connection = sqlite3.connect(database_pathname)

        self.share["database_pathname"] = database_pathname
        self.share["source_file"] = f"v{_VERSION}⇒ {__file__}"

    def test_command(self, parameter):
        print(f"Command: test_command({parameter})")

    def test_request(self, topic_path_response, request):
        aiko.message.publish(topic_path_response, "(item_count 1)")
        aiko.message.publish(topic_path_response, f"({request})")

# --------------------------------------------------------------------------- #

def do_command(actor_interface, command_handler, terminate=True):  # Refactor
    def actor_discovery_handler(command, service_details):
        if command == "add":
            event.remove_timer_handler(waiting_timer)
            topic_path = f"{service_details[0]}/in"
            actor = get_actor_mqtt(topic_path, actor_interface)
            command_handler(actor)
            if terminate:
                aiko.process.terminate()

    actor_discovery = ActorDiscovery(aiko.process)
    service_filter = ServiceFilter("*", "*", PROTOCOL, "*", "*", "*")
    actor_discovery.add_handler(actor_discovery_handler, service_filter)
    event.add_timer_handler(waiting_timer, 0.5)
    aiko.process.run()

item_count = 0
items_received = 0
response = []

def do_request(actor_interface, request_handler, response_handler):  # Refactor
    def topic_response_handler(_aiko, topic, payload_in):
        global item_count, items_received, response

        command, parameters = parse(payload_in)
        if command == "item_count" and len(parameters) == 1:
            item_count = int(parameters[0])
            items_received = 0
            response = []
        elif items_received < item_count:
            response.append((command, parameters))
            items_received += 1
            if items_received == item_count:
                response_handler(response)

    aiko.process.add_message_handler(topic_response_handler, _TOPIC_RESPONSE)
    do_command(actor_interface, request_handler, terminate=False)

def waiting_timer():
    event.remove_timer_handler(waiting_timer)
    print(f"Waiting for {ACTOR_TYPE}")

@click.group()

def main():
    pass

@main.command(help="Start Storage")
@click.argument("database_pathname", default="aiko_storage.db")

def start(database_pathname):
    tags = ["ec=true"]
    init_args = actor_args(ACTOR_TYPE, protocol=PROTOCOL, tags=tags)
    init_args["database_pathname"] = database_pathname
    storage = compose_instance(StorageImpl, init_args)
    storage.run()

@main.command(name="test_command")

def test_command():
    do_command(Storage, lambda storage: storage.test_command("hello"))

@main.command(name="test_request")
@click.argument("request")

def test_request(request):
    def response_handler(response):
        print(f"Response: {response}")
        import time; time.sleep(1)

    do_request(Storage,
        lambda storage: storage.test_request(_TOPIC_RESPONSE, request),
        response_handler
    )

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



from collections import defaultdict
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

import numpy as np
import shapely.geometry as geom
from pydantic import Field
from shapely.wkt import loads as wkt_loads

from highlighter.core.geometry import polygon_from_tlbr
from highlighter.core.hl_base_model import HasMany, HLModelMap

from ...core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    PIXEL_LOCATION_ATTRIBUTE_UUID,
    HLDataModel,
)
from .annotation import Annotation
from .datum_source import DatumSource
from .observation import Observation

__all__ = ["Entity"]


@HasMany("annotations", target_cls="highlighter.client.Annotation", back_populates="entity")
@HasMany("global_observations", target_cls="highlighter.client.Observation", back_populates="entity")
class Entity(HLDataModel):
    id: UUID = Field(..., default_factory=uuid4)
    annotations: HLModelMap = Field(..., default_factory=lambda: HLModelMap(Annotation))
    global_observations: HLModelMap = Field(..., default_factory=lambda: HLModelMap(Observation))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def make_entity(
        cls,
        entity_id: Optional[UUID] = None,
        location: Optional[geom.Polygon] = None,
        observations: List[Observation] = [],
        global_observations: List[Observation] = [],
    ):
        if entity_id is None:
            entity_id = uuid4()

        annotations = []
        if observations:
            annotations = [
                Annotation(
                    id=uuid4(),
                    entity_id=entity_id,
                    location=location,
                    observations=observations,
                    datum_source=observations[0].datum_source,
                )
            ]

        entity = cls(
            id=entity_id,
            annotations=annotations,
            global_observations=global_observations,
        )
        return entity

    def to_json(self):
        data = self.model_dump()
        data["id"] = str(data["id"])
        data["annotations"] = [a.to_json() for a in data["annotations"].values()]
        data["global_observations"] = [o.to_json() for o in data["global_observations"].values()]
        return data

    def serialize(self):
        # ToDo: Make this more general
        if self.annotations:
            annotations = [a.serialize() for a in self.annotations.values()]
        else:
            annotations = []

        if self.global_observations:
            global_observations = [o.serialize() for o in self.global_observations.values()]
        else:
            global_observations = []

        result = {
            "id": str(self.id),
            "annotations": annotations,
            "global_observations": global_observations,
        }
        return result

    def to_observations(self) -> List[Observation]:
        observations = []
        observations.extend(self.global_observations.values())
        for annotation in self.annotations:
            observations.extend(annotation.observations)
            if annotation.location is not None:
                observation = Observation.make_pixel_location_observation(
                    value=annotation.location,
                    confidence=annotation.datum_source.confidence,
                    occurred_at=datetime.now(timezone.utc),
                    frame_id=annotation.datum_source.frame_id,
                )
                observation.entity = annotation.entity
                observations.append(observation)
        return observations

    def to_deprecated_entity(self) -> "DeprecatedEntity":
        from haplomic.entity_attributes import Entity as DeprecatedEntity

        return DeprecatedEntity(
            entity_id=self.id,
            attributes={eavt.attribute_id: eavt for eavt in self.to_observations()},
        )

    @staticmethod
    def entities_from_assessment(assessment: "Submission") -> Dict[UUID, "Entity"]:
        # Step 1: Group annotations and observations by entity, and further group observations by annotation
        grouped_entities = defaultdict(
            lambda: {"annotations": [], "observations": defaultdict(list), "global_observations": []}
        )
        for annotation in assessment.annotations:
            grouped_entities[annotation.entity_id]["annotations"].append(annotation)
        for observation in assessment.entity_attribute_values:
            if observation.annotation_uuid is not None:
                grouped_entities[observation.entity_id]["observations"][observation.annotation_uuid].append(
                    observation
                )
            else:
                grouped_entities[observation.entity_id]["global_observations"].append(observation)
        # Step 2: Convert Hl Web representations into our local representations
        entities = {}
        for entity_id, entity_data in grouped_entities.items():
            entity = Entity(id=entity_id)
            for annotation in entity_data["annotations"]:
                ann = Annotation(
                    location=wkt_loads(annotation.location),
                    track_id=annotation.track_id,
                    data_file_id=annotation.data_file_id,
                    datum_source=DatumSource(
                        frame_id=annotation.frame_id,
                        confidence=annotation.confidence,
                    ),
                    correlation_id=annotation.correlation_id,
                )

                obs = Observation(
                    attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                    value=annotation.object_class.uuid,
                    occurred_at=datetime.now(
                        timezone.utc
                    ),  # TODO store with annotation in hl web or infer from source data file
                    datum_source=DatumSource(
                        frame_id=annotation.frame_id,
                        confidence=annotation.confidence,
                    ),
                )
                ann.observations.add(obs)
                entity.annotations.add(ann)

                for eav in entity_data["observations"][annotation.uuid]:
                    eavt = eav.to_eavt()
                    ann.observations.add(Observation.from_deprecated_eavt(eavt, uuid4()))

            for observation in entity_data["global_observations"]:
                entity.global_observations.add(
                    Observation(
                        attribute_id=observation.attribute_id,
                        value=observation.value,
                        occurred_at=observation.time,
                        datum_source=DatumSource(confidence=observation.entity_datum_source.confidence),
                    )
                )
            entities[entity_id] = entity
        return entities

    @staticmethod
    def frame_indexed_entities_from_avro(
        avro_entities, data_file_id: UUID
    ) -> List[Tuple[int, Dict[UUID, "Entity"]]]:
        """See Avro schema at highlighter.entity_avro_schema"""
        frame_indexed_entities = defaultdict(dict)  # Outer index is frame ID, inner index is entity ID
        for entity in avro_entities:
            # TODO handle embeddings
            # TODO handle eavts
            for track in entity.tracks:
                for detection in track.detections:
                    observations = [
                        Observation(
                            entity_id=entity.id,
                            attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                            value=entity.object_class,
                            occurred_at=datetime.now(timezone.utc),  # TODO change to correct occurred_at
                            datum_source=DatumSource(
                                confidence=1.0,
                                frame_id=detection.frame_id,
                            ),
                        )
                    ]
                    annotations = [
                        Annotation(
                            id=uuid4(),
                            entity_id=entity.id,
                            location=polygon_from_tlbr(detection.bounds),
                            track_id=track.track_id,
                            observations=observations,
                            data_file_id=data_file_id,
                            datum_source=DatumSource(
                                confidence=1.0,
                                frame_id=detection.frame_id,
                            ),
                        )
                    ]
                    global_observations = []
                    frame_indexed_entities[detection.frame_id][entity.id] = Entity(
                        id=entity.id,
                        annotations=annotations,
                        global_observations=global_observations,
                    )
        raise NotImplementedError("This implementation is a sketch, don't use without adding tests")
        return sorted(frame_indexed_entities.items(), key=lambda kv: kv[0])

    @staticmethod
    def entities_from_deprecated_eavts(eavts: List["EAVT"]) -> Dict[UUID, "Entity"]:
        entities = {}
        if len(eavts) > 0:
            grouped = defaultdict(list)
            for eavt in eavts:
                grouped[eavt.entity_id].append(eavt)
            for group in grouped.values():
                entity_id = group[0].entity_id
                location_eavts = [e for e in group if e.attribute_id == PIXEL_LOCATION_ATTRIBUTE_UUID]
                if len(location_eavts) == 0:
                    entities[entity_id] = Entity(id=entity_id)

                    for eavt in group:
                        obs_id = uuid4()
                        obs = Observation.from_deprecated_eavt(
                            eavt,
                            obs_id,
                        )

                        entities[entity_id].global_observations.add(obs)

                elif len(location_eavts) == 1:
                    location_eavt = location_eavts[0]
                    entity = Entity(id=entity_id)
                    annotation = Annotation(
                        id=uuid4(),
                        # The EAVT 'value' is a PixelLocationAttributeValue
                        # which has a shapely geometry as its 'value'
                        location=location_eavt.value.value,
                        datum_source=location_eavt.datum_source,
                    )

                    for eavt in group:
                        if eavt.attribute_id != PIXEL_LOCATION_ATTRIBUTE_UUID:
                            obs_id = uuid4()
                            obs = Observation.from_deprecated_eavt(
                                eavt,
                                obs_id,
                            )
                            annotation.observations.add(obs)

                    entity.annotations.add(annotation)
                    entities[entity_id] = entity
                else:
                    raise ValueError(
                        f"Can't handle {len(location_eavts)} pixel locations for a single entity"
                    )
        return entities

    @staticmethod
    def eavts_from_entities(entities: Dict[UUID, "Entity"]) -> List[Observation]:
        eavts = []
        for entity in entities.values():
            eavts.extend(entity.to_observations())
        return eavts

    @staticmethod
    def to_deprecated_entities(entities: Dict[UUID, "Entity"]) -> "DeprecatedEntities":
        from haplomic.entity_attributes import Entities as DeprecatedEntities

        deprecated_entities = DeprecatedEntities()
        for entity in entities.values():
            deprecated_entities.add_entity(entity.to_deprecated_entity())
        return deprecated_entities

    def get_observation_by_attribute_id(self, attribute_id: UUID) -> Optional[Observation]:
        for an in self.annotations:
            for obs in an.observations:
                if obs.attribute_id == attribute_id:
                    return obs
        return None

    def crop_annotations(
        self,
        crop_args: "CropArgs",
        crop_annotations_with: Optional[Union[UUID, Tuple[UUID, Any]]] = None,
    ) -> List[np.ndarray]:
        crops = []
        if crop_annotations_with is None:
            for a in self.annotations:
                crops.append(a.crop(crop_args))
        else:

            if isinstance(crop_annotations_with, UUID):
                value = None
                attribute_id = crop_annotations_with
            else:
                attribute_id, value = crop_annotations_with

            for a in self.annotations:
                if a.has_observation(attribute_id, value=value):
                    crops.append(a.crop(crop_args))

        return crops

    @staticmethod
    def crop_entity_annotations(
        entities: List["Entity"],
        crop_args: "CropArgs",
        crop_annotations_with: Optional[Union[UUID, Tuple[UUID, Any]]] = None,
    ) -> List[np.ndarray]:
        crops = []
        for e in entities:
            crops.extend(e.crop_annotations(crop_args, crop_annotations_with=crop_annotations_with))
        return crops



from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Literal, Optional, Tuple, Union
from uuid import UUID

import aiko_services as aiko
from aiko_services import PROTOCOL_PIPELINE
from aiko_services import DataSource as BaseDataSource
from aiko_services import (
    PipelineImpl,
    compose_instance,
    pipeline_args,
    pipeline_element_args,
)
from pydantic import BaseModel

from highlighter.core.exceptions import OptionalPackageMissingError

__all__ = [
    "Capability",
    "DataSourceCapability",
    "ContextPipelineElement",
    "EntityUUID",
    "PROTOCOL_PIPELINE",
    "PipelineElement",
    "PipelineImpl",
    "StreamEvent",
    "compose_instance",
    "compose_instance",
    "pipeline_args",
    "pipeline_element_args",
]

EntityUUID = UUID

"""Decouple the rest of the code from aiko.PipelineElement"""
ContextPipelineElement = aiko.ContextPipelineElement
StreamEvent = aiko.StreamEvent
PipelineElement = aiko.PipelineElement

# SEPARATOR = b"\x1c"  # ASCII 28 (File Separator)
SEPARATOR = 28  # ASCII 28 (File Separator)


class _BaseCapability:
    class DefaultStreamParameters(BaseModel):
        """Populate with default stream param key fields"""

        pass

    @classmethod
    def default_stream_parameters(cls) -> BaseModel:
        return {
            k: v.default for k, v in cls.DefaultStreamParameters.model_fields.items() if not v.is_required()
        }

    def _get_parameter(
        self, name, default=None, required=False, use_pipeline=True, self_share_priority=True
    ) -> Tuple[Any, bool]:
        """Adds the correct output type to get_parameter type checking
        does not complain
        """
        return self.get_parameter(
            name,
            default=default,
            required=required,
            use_pipeline=use_pipeline,
            self_share_priority=self_share_priority,
        )


class Capability(PipelineElement, _BaseCapability):

    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, *args) -> Tuple[StreamEvent, Optional[Dict]]:
        raise NotImplementedError()

    def start_stream(self, stream, stream_id, use_create_frame=True):
        validated_parameters = self.DefaultStreamParameters(**self.parameters)
        for param_name in self.DefaultStreamParameters.model_fields:
            self.parameters[f"{self.definition.name}.{param_name}"] = getattr(
                validated_parameters, param_name
            )
            self.parameters.pop(param_name, None)
        stream.parameters.update(self.parameters)
        return StreamEvent.OKAY, {}


class DataSourceType(BaseModel):
    # class MediaType(str, Enum):
    #    IMAGE = "IMAGE"
    #    TEXT = "TEXT"
    #    VIDEO = "VIDEO"

    media_type: str
    url: str
    id: UUID
    content: Optional[Any] = None

    @classmethod
    def image_iter(cls, images: Iterable[Union[str, Path, bytes]]):
        pass

    @classmethod
    def video_iter(cls, videos: Iterable[Union[str, Path, bytes]]):
        pass

    @classmethod
    def text_iter(cls, tests: Iterable[Union[str, Path, bytes]]):
        pass


class DataSourceCapability(BaseDataSource, _BaseCapability):

    stream_media_type = None

    class DefaultStreamParameters(_BaseCapability.DefaultStreamParameters):

        rate: Optional[float] = None
        batch_size: int = 1
        data_sources: List[DataSourceType] = []
        file_ids: Optional[Iterable] = None
        task_id: Optional[UUID] = None

    @property
    def rate(self) -> float:
        return self._get_parameter("rate")[0]

    @property
    def batch_size(self) -> int:
        return self._get_parameter("batch_size")[0]

    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    def frame_batch_generator(self, stream, pipeline_iter_idx):
        """Produce a batch of frames.

        Args:
            stream: The Stream context
            pipeline_iter_idx: An integer counting the number of times the
                               pipeline has been executed, (ie: process_frame
                               has been called)

        """

        batch_size = self.batch_size
        task_id, _ = self._get_parameter("task_id")

        frame_data_batch = defaultdict(list)
        for _ in range(batch_size):
            frame_data_gen = stream.variables["frame_data_generator"]
            try:
                frame_data = next(frame_data_gen)
            except StopIteration:
                break
            except Exception as e:
                return StreamEvent.ERROR, {"diagnostic": e}

            self.logger.debug(f"frame_data: {frame_data}")

            # Convert list of dicts to dict or lists
            # [{"data_files": ..., "entities": ...}, ...]
            # {"data_files": [...], "entities": [...]}
            for k, v in frame_data.items():
                frame_data_batch[k].append(v)

        # For each pipeline iteration the is a batch of file_ids and frame_ids
        stream.variables["task_id"] = task_id

        if len(frame_data_batch):
            return StreamEvent.OKAY, frame_data_batch
        else:
            return StreamEvent.STOP, {"diagnostic": "All frames generated"}

    def frame_data_generator(self, data_sources):
        pass

    def start_stream(self, stream, stream_id, use_create_frame=True):
        data_sources, _ = self._get_parameter("data_sources")

        if not data_sources:
            raise ValueError("data_sources is empty")

        if getattr(self, "stream_media_type", None) is None:
            raise ValueError(
                "All subclasses of DataSourceCapability, must have "
                "'stream_media_type' set as a class variable"
            )

        stream.variables["stream_media_type"] = self.stream_media_type
        stream.variables["frame_data_generator"] = self.frame_data_generator(data_sources)

        self.create_frames(stream, self.frame_batch_generator, rate=self.rate)
        return StreamEvent.OKAY, {}

    def process_frame(self, stream, data_files: List[str], entities) -> Tuple[StreamEvent, Dict]:
        return StreamEvent.OKAY, {"data_files": data_files, "entities": entities}

    def stop_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        self.stop()
        return StreamEvent.OKAY, {}
